function _extends() { _extends = Object.assign || function (target) { for (var i = 1; i < arguments.length; i++) { var source = arguments[i]; for (var key in source) { if (Object.prototype.hasOwnProperty.call(source, key)) { target[key] = source[key]; } } } return target; }; return _extends.apply(this, arguments); }

function _asyncToGenerator(fn) { return function () { var self = this, args = arguments; return new Promise(function (resolve, reject) { var gen = fn.apply(self, args); function step(key, arg) { try { var info = gen[key](arg); var value = info.value; } catch (error) { reject(error); return; } if (info.done) { resolve(value); } else { Promise.resolve(value).then(_next, _throw); } } function _next(value) { step("next", value); } function _throw(err) { step("throw", err); } _next(); }); }; }

const config = require('./config');

const {
  asyncify,
  queue
} = require('async');

const {
  deserializeCommit
} = require('./dynamodb');

class Projector {
  constructor(projections, {
    tableName = config.tableName,
    logger = config.logger,
    pollInterval = 500
  } = {}) {
    Object.assign(this, {
      pollInterval,
      projections,
      tableName,
      projectionCommitPromises: {},
      bufferedStreamCommits: [],
      running: false,
      maxQueueBufferSize: 1000,
      minNumberOfCommitsPerQuery: 100,
      maxConcurrency: Math.min(Object.values(projections).map(p => p.maxConcurrency)),
      logger,
      commitBuffer: [],
      dynamodb: new config.configuredAWS.DynamoDB(),
      exclusiveAggregateTypes: Object.values(projections).reduce((types, projection) => [...types, ...projection.exclusiveAggregateTypes.filter(type => !types.includes(type))], [])
    });
  }

  start({
    poll = true
  }) {
    var _this = this;

    return _asyncToGenerator(function* () {
      _this.logger.info(`Projector started with poll interval of ${_this.pollInterval}ms`);

      if (_this.running) throw new Error('Already running');
      _this.running = true;
      const projections = Object.values(_this.projections); // Verify that all projections are in sync

      const statusList = yield Promise.all(projections.map(projection => projection.readProjectionStatus()));

      if (statusList.some(({
        commitsInProgress
      }) => commitsInProgress) || new Set(statusList.map(status => status.headCommitId)).size > 1) {
        _this.logger.warn('A projection was dirty, clearing all projections and rebuilding..');

        yield Promise.all(projections.map(p => p.clearProjection()));
        return _this.start({
          poll
        });
      }

      _this.headCommitId = statusList[0].headCommitId;

      if (_this.headCommitId) {
        _this.logger.debug(`Projections are at commitId '${_this.headCommitId}'`);
      } else {
        _this.logger.debug('Projections are empty');
      }

      while (_this.running) {
        yield _this.hydrateCommitQueue();

        const commitsToProcess = _this.checkoutParallelizableCommitsFromBuffer();

        if (!commitsToProcess.length) {
          yield new Promise(resolve => setTimeout(resolve, _this.pollInterval));
          continue;
        }

        try {
          yield Promise.all([_this.writeProjectionStatus({
            headCommitId: _this.headCommitId,
            processing: true
          }), _this.processCommits(commitsToProcess)]);
        } catch (error) {
          console.log(error);
          process.exit(1);
        }

        _this.headCommitId = commitsToProcess[commitsToProcess.length - 1].commitId;
        yield _this.writeProjectionStatus({
          headCommitId: _this.headCommitId,
          processing: false
        });

        _this.logger.debug(`After ${commitsToProcess.length} commits, projections are now at commitId '${_this.headCommitId}'`);
      }
    })();
  }

  writeProjectionStatus(status) {
    return Promise.all(Object.values(this.projections).map(projection => projection.writeProjectionStatus(status)));
  }

  stop() {// return this.subscriber.stop()

    return _asyncToGenerator(function* () {})();
  }

  fetchCommits({
    limit = 1000
  }) {
    var _this2 = this;

    return _asyncToGenerator(function* () {
      let queryResult = {};
      const commits = [];
      const currentCommitId = _this2.commitBuffer.length ? _this2.commitBuffer[_this2.commitBuffer.length - 1].commitId : _this2.headCommitId || '0';

      do {
        queryResult = yield _this2.dynamodb.query(_extends({
          TableName: _this2.tableName,
          IndexName: 'commitIdIndex',
          KeyConditionExpression: 'z = :a AND c > :c',
          ExpressionAttributeValues: {
            ':a': {
              S: 't'
            },
            ':c': {
              S: currentCommitId
            }
          },
          Limit: limit - commits.length
        }, queryResult.LastEvaluatedKey && {
          ExclusiveStartKey: queryResult.LastEvaluatedKey
        })).promise();
        queryResult.Items.map(deserializeCommit).forEach(commit => commits.push(commit));
      } while (queryResult.LastEvaluatedKey && commits.length < limit);

      return commits;
    })();
  }

  hydrateCommitQueue() {
    var _this3 = this;

    return _asyncToGenerator(function* () {
      const numberOfCommitsToFetch = _this3.maxQueueBufferSize - _this3.commitBuffer.length;

      if (numberOfCommitsToFetch > _this3.minNumberOfCommitsPerQuery) {
        const commits = yield _this3.fetchCommits({
          limit: numberOfCommitsToFetch
        });
        if (commits.length) _this3.logger.debug(`Got ${commits.length} commits`);
        commits.forEach(commit => _this3.commitBuffer.push(commit));
      }
    })();
  }

  checkoutParallelizableCommitsFromBuffer() {
    const commitList = [];
    const aggregateIdMap = {};

    for (const [i, commit] of this.commitBuffer.entries()) {
      const {
        aggregateType,
        aggregateKey
      } = commit;

      if (aggregateIdMap[`${aggregateType}:${aggregateKey}`]) {
        // We need to wait for the previous commit for this aggregateId to finish
        continue; // Check if there are other commits we can include
      }

      if (this.exclusiveAggregateTypes.includes(aggregateType)) {
        // This commit needs to be handled exlusively
        if (!commitList.length) commitList.push([i, commit]); // Fine, we are alone

        break; // No more commits
      } // Include commit in commitList


      aggregateIdMap[`${aggregateType}:${aggregateKey}`] = true;
      commitList.push([i, commit]);
    }

    for (const [deleteCount, [initialCommitIndex]] of commitList.entries()) {
      this.commitBuffer.splice(initialCommitIndex - deleteCount, 1);
    }

    return commitList.map(([, commit]) => commit);
  }

  processCommits(commits) {
    this.logger.debug(`Process commits (${commits.length})`);
    return new Promise((resolve, reject) => {
      const processingQueue = queue(asyncify(commit => Promise.all(Object.values(this.projections).map(projection => projection.processCommit(commit)))), this.maxConcurrency);
      processingQueue.error = reject;
      processingQueue.drain = resolve;

      for (const commit of commits) {
        processingQueue.push(commit);
      }
    });
  }

}

module.exports = Projector;