function _extends() { _extends = Object.assign || function (target) { for (var i = 1; i < arguments.length; i++) { var source = arguments[i]; for (var key in source) { if (Object.prototype.hasOwnProperty.call(source, key)) { target[key] = source[key]; } } } return target; }; return _extends.apply(this, arguments); }

function _objectWithoutProperties(source, excluded) { if (source == null) return {}; var target = {}; var sourceKeys = Object.keys(source); var key, i; for (i = 0; i < sourceKeys.length; i++) { key = sourceKeys[i]; if (excluded.indexOf(key) >= 0) continue; target[key] = source[key]; } if (Object.getOwnPropertySymbols) { var sourceSymbolKeys = Object.getOwnPropertySymbols(source); for (i = 0; i < sourceSymbolKeys.length; i++) { key = sourceSymbolKeys[i]; if (excluded.indexOf(key) >= 0) continue; if (!Object.prototype.propertyIsEnumerable.call(source, key)) continue; target[key] = source[key]; } } return target; }

function _asyncToGenerator(fn) { return function () { var self = this, args = arguments; return new Promise(function (resolve, reject) { var gen = fn.apply(self, args); function step(key, arg) { try { var info = gen[key](arg); var value = info.value; } catch (error) { reject(error); return; } if (info.done) { resolve(value); } else { Promise.resolve(value).then(_next, _throw); } } function _next(value) { step("next", value); } function _throw(err) { step("throw", err); } _next(); }); }; }

const WebSocket = require('ws');

const config = require('../config');

const {
  deserializeCommit
} = require('../dynamodb');

const eventMatchesSubscriptions = require('./eventMatchesSubscriptions');

class EventStreamServer {
  constructor({
    verifyClient,
    port = 80,
    pollDelay = 500,
    tableName = config.tableName,
    dynamodb = new config.configuredAWS.DynamoDB()
  } = {}) {
    this.logger = config.logger;
    this.server = new WebSocket.Server({
      verifyClient,
      port
    });
    this.logger.info(`EventStreamServer: Listening on port ${port}`);
    this.server.on('error', error => this.handleError(error));
    this.server.on('connection', this.handleConnection.bind(this));
    this.server.on('close', this.handleConnection.bind(this));
    Object.assign(this, {
      pollDelay,
      tableName,
      dynamodb,
      isPolling: false
    });
  }

  handleError(error) {
    this.logger.error(error);
  }

  handleConnection(client) {
    const clientAddress = client._socket.remoteAddress;
    this.logger.debug(`EventStreamServer: new client (${clientAddress})`);
    if (!this.isPolling) this.startPolling();
    client.filterSets = [];
    client.on('message', json => {
      this.logger.debug(`EventStreamServer: filter sets for ${clientAddress} set to ${json}`);
      client.filterSets = JSON.parse(json);
    });
    client.on('close', () => {
      this.logger.debug(`EventStreamServer: client disconnected (${clientAddress})`);
    });
  }

  close() {
    this.server.close();
    this.server = undefined;
  }

  startPolling() {
    var _this = this;

    return _asyncToGenerator(function* () {
      _this.logger.debug(`EventStreamServer: we have clients, started polling (delay: ${_this.pollDelay}ms)`);

      _this.isPolling = true;
      let commitIdCursor = new Date().toISOString().replace(/[^0-9]/g, '');

      while (_this.server && _this.server.clients.size) {
        const queryResult = yield _this.dynamodb.query({
          TableName: _this.tableName,
          IndexName: 'commitIdIndex',
          KeyConditionExpression: 'z = :z AND c > :c',
          ExpressionAttributeValues: {
            ':z': {
              S: 't'
            },
            ':c': {
              S: commitIdCursor
            }
          }
        }).promise();

        if (!queryResult.Items.length) {
          yield new Promise(resolve => setTimeout(resolve, _this.pollDelay));
          continue;
        }

        const commits = queryResult.Items.map(deserializeCommit);
        let eventsSentCount = 0;

        for (const _ref of commits) {
          const {
            events
          } = _ref,
                commitRest = _objectWithoutProperties(_ref, ["events"]);

          for (const event of events) {
            const decoratedEvent = _extends({}, event, commitRest);

            let matches = 0;

            for (const client of _this.server.clients) {
              if (eventMatchesSubscriptions(decoratedEvent, client.filterSets)) {
                matches++;
                client.send(JSON.stringify(decoratedEvent));
                eventsSentCount++;
              }
            }

            _this.logger.trace(`EventStreamServer: [${decoratedEvent.type}] sent to ${matches} clients`);
          }

          commitIdCursor = commitRest.commitId;
        }

        _this.logger.debug(`EventStreamServer: processed ${commits.length} commits and sent ${eventsSentCount} events to clients`);
      }

      _this.isPolling = false;

      _this.logger.debug('EventStreamServer: no clients, stopping polling');
    })();
  }

}

module.exports = EventStreamServer;